package net.knarcraft.minecraftserverlauncher.server;

import net.knarcraft.minecraftserverlauncher.server.servertypes.*;

import javax.naming.ConfigurationException;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Scanner;

import static net.knarcraft.minecraftserverlauncher.utility.CommonFunctions.getResourceAsScanner;

/**
 * A class which keeps track of available server types
 */
public class ServerTypeHandler {

    private static final ArrayList<ServerType> serverTypes = new ArrayList<>();

    /**
     * Gets a list of all server types' names
     *
     * @return <p>A list containing the names of all server types</p>
     */
    public static String[] getTypeNames() throws ConfigurationException {
        if (serverTypes.isEmpty()) {
            loadServerTypes();
        }
        ArrayList<ServerType> types = getServerTypes();
        String[] serverTypeNames = new String[types.size()];
        for (int i = 0; i < types.size(); i++) {
            serverTypeNames[i] = types.get(i).getName();
        }
        return serverTypeNames;
    }

    /**
     * Gets all instantiated server types
     *
     * @return <p>A list of server types</p>
     */
    public static ArrayList<ServerType> getServerTypes() throws ConfigurationException {
        if (serverTypes.isEmpty()) {
            loadServerTypes();
        }
        return serverTypes;
    }

    /**
     * Gets a server type by the given name
     *
     * @param name <p>Then name of the server type</p>
     * @return <p>A AbstractServerType</p>
     */
    public static ServerType getByName(String name) throws ConfigurationException {
        if (serverTypes.isEmpty()) {
            loadServerTypes();
        }
        for (ServerType serverType : serverTypes) {
            if (serverType.getName().equals(name)) {
                return serverType;
            }
        }
        return null;
    }

    /**
     * Reads valid server types and version from a file, and creates their objects
     *
     * @throws ConfigurationException <p>If anything goes wrong</p>
     */
    private static void loadServerTypes() throws ConfigurationException {
        Scanner file;
        try {
            file = getResourceAsScanner("servertypes.csv");
        } catch (FileNotFoundException e) {
            throw new ConfigurationException("Server type configuration file is missing.");
        }
        while (file.hasNextLine()) {
            //Splits the next file line into arguments
            String[] serverTypeInfo = file.nextLine().split(";", -1);

            //Gets list of server versions from file line
            String[] serverVersions;
            if (serverTypeInfo[1].contains(",")) {
                serverVersions = serverTypeInfo[1].split(",", -1);
            } else {
                serverVersions = new String[]{serverTypeInfo[1]};
            }

            addServerType(serverTypeInfo, serverVersions);
        }
    }

    /**
     * Adds a new server type
     *
     * @param serverTypeInfo <p>A list containing necessary information for initializing the server type</p>
     * @param serverVersions <p>A list of all server versions for the server type</p>
     */
    private static void addServerType(String[] serverTypeInfo, String[] serverVersions) {
        ServerType newType;
        switch (serverTypeInfo[0]) {
            case "Craftbukkit":
                newType = new CraftBukkit("Bukkit", false, serverVersions, serverTypeInfo[2],
                        serverTypeInfo[3]);
                break;
            case "Spigot":
                newType = new Spigot("Spigot", false, serverVersions, serverTypeInfo[2],
                        serverTypeInfo[3]);
                break;
            case "Vanilla":
                newType = new Vanilla("Vanilla", false, serverVersions, serverTypeInfo[2],
                        serverTypeInfo[3]);
                break;
            case "MCPCplus":
                newType = new MCPCPlus("MCPCplus", false, serverVersions, serverTypeInfo[2],
                        serverTypeInfo[3]);
                break;
            case "Paper":
                newType = new Paper("Paper", false, serverVersions, serverTypeInfo[2],
                        serverTypeInfo[3]);
                break;
            case "Bungee":
                newType = new BungeeCord("Bungee", true, serverVersions, serverTypeInfo[2],
                        serverTypeInfo[3], serverTypeInfo[4], serverTypeInfo[5], serverTypeInfo[6], serverTypeInfo[7]);
                break;
            case "Travertine":
                newType = new Travertine("Travertine", true, serverVersions, serverTypeInfo[2],
                        serverTypeInfo[3], serverTypeInfo[4], serverTypeInfo[5]);
                break;
            case "Waterfall":
                newType = new Waterfall("Waterfall", true, serverVersions, serverTypeInfo[2],
                        serverTypeInfo[3], serverTypeInfo[4], serverTypeInfo[5]);
                break;
            case "SpongeVanilla":
                newType = new SpongeVanilla("SpongeVanilla", false, serverVersions,
                        serverTypeInfo[2], serverTypeInfo[3], serverTypeInfo[4], serverTypeInfo[5], serverTypeInfo[6]);
                break;
            case "SpongeForge":
                newType = new SpongeForge("SpongeForge", false, serverVersions,
                        serverTypeInfo[2], serverTypeInfo[3], serverTypeInfo[4], serverTypeInfo[5], serverTypeInfo[6]);
                break;
            case "Custom":
                newType = new Custom("Custom");
                break;
            default:
                throw new IllegalArgumentException("Unknown server type defined in config file.");
        }
        serverTypes.add(newType);
    }
}
